open String

let version = "pftdbns, version 0.4.2 Tue Apr 24 16:08:35 CEST 2007"


let recordlist_r = ref []
let filenames = Hashtbl.create 10000
let filtered = Hashtbl.create 10000

type function_selection_t = Move | List | List_and_structure | List_and_directory | List_and_all | Not_selected
type dirname_options_t    = Standard | Prefix of string | Fixed of string
type names_t              = { filename: string; structure: string; basename: string; dirname: string; newdirname: string }
(* filename: path of file (complete name) ; basename: basename; dirname: path to file *)
type filter_selection_t   = Normal | Inverted
type filekind_selection_t = Disallow_dirs | Allow_dirs


let selection        = ref Not_selected
let dirnameopt       = ref Standard
let filter_selection = ref Normal
let fk_sel           = ref Disallow_dirs


(*

     pftdbns: put files to directories (sorted) by name structure
    ==============================================================

     Use at your own risk, no guarantee provided.

  What this program is good for: Sorting files
  into directories. The sorting algorithm of the filenames is based 
  on categorizing the filenames into character-classes.
  
  The character-classes are groups of files that are
  based upon the lexical structure of the filenmaes.

  But what do I mean with "lexical structure"?
  Well, maybe you can argue if "lexical structure"
  is the right name for that, but I mean the following:
  You can sort not only by the characters of a string.
  You also can sort by character-classes of the caracters
  of the string.
  That means sorting by a string it's character-class-structure.

  See explanation and examples in the file "README".

  This program is written by Oliver Bandel (Berlin, Germany),
  all rights reserved. (Copyright/Urheberrecht.)
  But I allow you to redistribute this software under the
  Terms of the GPL (version 2). See file "COPYING" for more
  details.

*)


(* ------------------------------------------------------------------------- *)
(* The function "scan_string" scans the string and gives back a string which *)
(* contains the structure of the input-string, where structure means, that   *)
(* all characters of the string will be seen as part of a character class in *)
(* which they will be "sorted" in.                                           *)
(* "scan_string" could have been also be called "classify_string".           *)
(* ------------------------------------------------------------------------- *)
let scan_string str =
   let str = String.copy str in (* overwrite name in this environment *)
     let chartype chr = match chr with
         'a' .. 'z' | 'A' .. 'Z' -> 'l'
       | '0' .. '9'              -> 'i'
       | ' ' | '\t' | '\n'       -> 'b'
       | '.'                     -> 'd'
       | '/'                     -> 's'
       | _                       -> 'o'
     in
       for i = 0 to (String.length str - 1) 
         do str.[i] <- chartype str.[i] done; str (* replace each char by its charclass-id *)

(* ------------------------------------------------------------------------- *)
(* This is for cascading filters.                                            *)
(* ------------------------------------------------------------------------- *)
(* If you have more than one filter to apply to an element, you can use this *)
(* filter-cascade creating function. This is extremely useful if you have    *)
(* to filter data with many filters, and especially, if you have to use many *)
(* different cascades in your code.                                          *)
(* You not have to hand code the whole filter-pipeline often.                *)
(* You only define your filters and then combine them with this function and *)
(* specify all used filters as a list, giving this list as argument to the   *)
(* "filt_cascade" function. :)                                               *)
(* ------------------------------------------------------------------------- *)
let filt_cascade filterlist element =
   let rec fc_aux bool_sofar filtlist =
    match filtlist with
     [] -> bool_sofar
   | hd::tl -> fc_aux (bool_sofar & hd element) tl
   in
     fc_aux true filterlist
(*
 (*
 val filt_cascade : ('a -> bool) list -> 'a -> bool = <fun>
 *)
 # let f_x = filt_cascade [f_1;f_2;f_3];;
*)


(* -------------------------------------------------------------------- *)
(* gives back a boolean to decide if "name" is name of directory or not *)
(* -------------------------------------------------------------------- *)
let is_dir name = let status = Unix.stat name in
      match status.Unix.st_kind with Unix.S_DIR -> true | _ -> false

let is_regfile name = let status = Unix.stat name in
      match status.Unix.st_kind with Unix.S_REG -> true | _ -> false


(* --------------- *)
(* comparing names *)
(* --------------- *)
module NewCompare =
  struct

  (* shorter files behave *always* like a character coming first in the alphabet so *)
  (* "z" comes before "aa", but "a" comes before "z" and "ab" comes after "aa".     *)

  let compare_pair a b =   (* "overrides" Pervasives.compare in this environment (new name-val-binding) *)
      let fa = fst a
      and fb = fst b
      in
        let diff = String.length fa - String.length fb
        in
          if    diff = 0
           then Pervasives.compare fa fb (* depending on the string-contents      *)
           else diff/(abs diff)          (* -1 or +1, depending on string-lengths *)


  let sort_records_by_structure recordlist = 
     let lclist = List.rev_map (fun record -> ((String.lowercase (record.filename)), record)) recordlist
     in
       let sorted = List.sort compare_pair lclist
       in
         List.rev (List.rev_map (fun (lowercase,record) -> record) sorted)

  end


(* ========================================================================= *)


(*
    ANDERE VARIANTE:
       Hash nehmen und String-Struktur als key nehmen und alle Filenames einspeichern
       und dann nach keys sortiert die Ergebnisse mit normalem Sort sortieren.
       => ggf. alle Files mit selber Namens-Struktur in ein eigenes Verzeichnis ablegen. (!)
       (fr letzteres kann das Sortieren der Files mit gleichstrukturiertem Namen entfallen.
*)


(* ------------------------------------------------------------------- *)
(* rename (move) file to dir with the name of the filename's structure *)
(* ------------------------------------------------------------------- *)
(* But do not overwrite existing files!                                *)
(* ------------------------------------------------------------------- *)
let mymove oldname newname =
  if Sys.file_exists newname
  then prerr_endline ("ignoring file " ^ oldname ^ " to prevent overwriting an existing file!")
  else
    begin
      try
        Unix.rename oldname newname
      with
        | Unix.Unix_error (error,_,_)  -> Printf.printf "%s: oldname: \"%s\" newname: \"%s\"\n"  (Unix.error_message error) oldname newname
    end


(* ----------------------------------------------------------------- *)
(* Move the file into a directory with the name of file's structure  *)
(* ----------------------------------------------------------------- *)
let move_to_newdir record =
      if Sys.file_exists record.newdirname = false
      then Unix.mkdir record.newdirname 448 else(); (* ***** IF file_exists, but is not a directory => IGNORE Action !!! ***** *)
      mymove record.filename (record.newdirname ^ "/" ^ record.basename)

(* -------------------------------------------------------- *)
(* print usage-message, when there are no (valid) filenames *)
(* -------------------------------------------------------- *)
let print_message () = print_endline "you gave no valid filenames to me!";
                       print_endline "usage: please, give filenames as argument to me."

let print_message_and_exit () = print_message(); exit 1


let parse () = (*print_endline "OK, starting the parse!";*)
    Arg.parse [
         ("-dp", Arg.String (fun x -> dirnameopt := Prefix x), " <dirpref> directory name-prefix");
         ("-fd", Arg.String (fun x -> dirnameopt := Fixed x), "<fixdir> fixed directory name");
         ("-l", Arg.Unit (fun () -> selection := List), " function selection: only list filenames; do not move files" );
         ("-ls", Arg.Unit (fun () -> selection := List_and_structure), "function selection: only list filenames and name-structure; do not move files" );
         ("-ld", Arg.Unit (fun () -> selection := List_and_directory), "function selection: only list filenames and directory name; do not move files" );
         ("-la", Arg.Unit (fun () -> selection := List_and_all), "function selection: only list (list all); do not move files" );
         ("-tf", Arg.String (fun x -> if Sys.file_exists x then Hashtbl.replace filtered x ""), "<templatename> template filter");
         ("-inv", Arg.Unit (fun () -> filter_selection := Inverted ), " invert the template filter (-tf): use NON-matching files of -tf <template>" );
         ("-ad", Arg.Unit (fun () -> fk_sel := Allow_dirs ), " allow dirs to be moved (and regarded as files in general) usage DEPRECATED!" );
         ("-v", Arg.Unit (fun () -> Printf.printf "%s\n" version;exit 0), " version of the program" )
              ] 
         ( fun name -> if Sys.file_exists name then Hashtbl.replace filenames name "" ) 
         "Use the program with following options:"


(* ------------------------ *)
(* Main-Part of the Program *)
(* ------------------------ *)

open NewCompare


let _ = 

    parse(); (* CLI-parsing *)
    let create_record = 
        begin
          match !dirnameopt with
            | Standard        -> fun fn -> let ba = Filename.basename fn in
                                           let dn = Filename.dirname fn in
                                           let stru = scan_string ba in
                                             {filename = fn; basename = ba; structure = stru; dirname = dn; newdirname = dn ^ "/" ^  stru}
            | Prefix dirpref  -> fun fn -> let ba = Filename.basename fn in
                                           let dn = Filename.dirname fn in
                                           let stru = scan_string ba in
                                             {filename = fn; basename = ba; structure = stru; dirname = dn ; newdirname = dn ^ "/" ^ dirpref ^ stru }
            | Fixed  dirnam   -> fun fn -> let ba = Filename.basename fn in
                                           let dn = Filename.dirname fn in
                                           let stru = scan_string ba in
                                             {filename = fn; basename = ba; structure = stru; dirname = dn; newdirname = dirnam}
        end
        in

        (* if there are filters enabled (-tf opption used) then filter filenames... *)
        if Hashtbl.length filtered > 0 then
        begin
        (*print_endline "***filtering is enabled! ***";*)
          match !filter_selection with
              (* Normal mode: if filtered, then  files are OK to use => if not in filter list, then throw them away *)
              Normal   -> Hashtbl.iter (fun k v -> if  not (Hashtbl.mem filtered k) then Hashtbl.remove filenames k) (Hashtbl.copy filenames)
              (* Inverted mode: if filtered, then files are to IGNORE; this means: if in filter list, then throw them away *)
            | Inverted -> Hashtbl.iter (fun k v -> if  (Hashtbl.mem filtered k)  then Hashtbl.remove filenames  k) filtered
        end
      ;

      (*

         If Dirs are NOT allowed:

                        file is_dir  ||  remove from list
                       --------------++------------------
                             0       ||        0          
                             1       ||        1          

      *)

      (* If Directories are not allowed, throw them away *)
      if not (!fk_sel = Allow_dirs)
      then Hashtbl.iter (fun k v -> if is_dir k  then Hashtbl.remove filenames k) (Hashtbl.copy filenames);

      if Hashtbl.length filenames = 0 then print_message_and_exit () else ();

      Hashtbl.iter (fun k v -> recordlist_r := k :: !recordlist_r) filenames ;


         let sorted_records = sort_records_by_structure (List.map create_record !recordlist_r)
           in
            match !selection with
              | Move | Not_selected -> List.iter move_to_newdir sorted_records (* DEFAULT: MOVE *)
              | List                -> List.iter (fun record -> Printf.printf "%s\n" record.filename) sorted_records
              | List_and_structure  -> List.iter (fun r -> Printf.printf "%s => %s\n" r.filename r.structure ) sorted_records
              | List_and_directory  -> List.iter (fun r -> Printf.printf "%s => %s\n" r.filename r.newdirname ) sorted_records
              | List_and_all        -> List.iter (fun r -> Printf.printf "%s => %s => %s\n" r.filename r.structure r.newdirname ) sorted_records

